from __future__ import annotations

import math
import os
from typing import TYPE_CHECKING, Any, Dict, Final, List, Optional, Tuple, cast
import uuid

from google.protobuf.timestamp_pb2 import Timestamp
import grpc

from jetpack import __version__
from jetpack._runtime.errors import JetpackError, RuntimeError
from jetpack._runtime.interceptor import header_adder_interceptor
from jetpack._runtime.local import LocalStub
from jetpack._util import jetpickle
from jetpack._util.network import codec
from jetpack.config import instrumentation, k8s, namespace
from jetpack.proto.runtime.v1alpha1 import remote_pb2, remote_pb2_grpc

# Instead of having individual clients per functionality (e.g. job), have a
# single client for our runtime.

# TODO(Landau): Move client functionality out of _job.client and centralize it
# here.

# TODO(landau): make this a param
MAX_MESSAGE_LENGTH: Final[int] = 10485760  # 10MB

# Prevent circular dependency
if TYPE_CHECKING:
    from jetpack._task.task import Task


class Client:
    def __init__(self) -> None:
        self.host = os.environ.get("JETPACK_RUNTIME_SERVICE_HOST", "")
        self.api_key = ""
        self.stub: Optional[remote_pb2_grpc.RemoteExecutorStub] = None
        pass

    def set_runtime_host(self, host: str, api_key: str) -> None:
        self.host = host
        self.api_key = api_key

    def _address(self) -> str:
        default_port = "443"
        if self.api_key == "":
            default_port = "80"
        port: str = os.environ.get("JETPACK_RUNTIME_SERVICE_PORT", default_port)
        address = f"{self.host.strip()}:{port.strip()}"
        return address

    def dial(self) -> remote_pb2_grpc.RemoteExecutorStub:
        if self.stub is None:
            address = self._address()
            if self._has_runtime_credentials():
                sdk_info_interceptor = header_adder_interceptor(
                    "x-jetpack-sdk-version", f"python-{__version__}"
                )
                options = [
                    ("grpc.max_send_message_length", MAX_MESSAGE_LENGTH),
                ]
                interceptors = [
                    sdk_info_interceptor,
                ]
                if self.api_key == "":
                    channel = grpc.aio.insecure_channel(
                        address,
                        options=options,
                        interceptors=interceptors,
                    )
                else:
                    # https://grpc.github.io/grpc/python/grpc.html#create-client-credentials
                    creds = grpc.composite_channel_credentials(
                        grpc.ssl_channel_credentials(),
                        # This gets sent as an authorization bearer header:
                        grpc.access_token_call_credentials(f"{self.api_key}"),
                    )

                    channel = grpc.aio.secure_channel(
                        address,
                        creds,
                        options=options,
                        interceptors=interceptors,
                    )
                self.stub = remote_pb2_grpc.RemoteExecutorStub(channel)
            else:
                self.stub = LocalStub(self)

        return self.stub

    def _has_runtime_credentials(self) -> bool:
        # implicit credentials: is in cluster
        # explicit credentials: api key
        return k8s.is_in_cluster() or self.api_key != ""

    async def register_app(
        self,
        cron_jobs: List[remote_pb2.CronJob],
    ) -> remote_pb2.RegisterAppResponse:
        request = remote_pb2.RegisterAppRequest(
            namespace=cast(str, namespace.get()),
            hostname=os.environ["HOSTNAME"],
            cron_jobs=cron_jobs,
        )
        stub = self.dial()
        response: remote_pb2.RegisterAppResponse = await stub.RegisterApp(request)
        return response

    async def create_task(
        self,
        task: Task,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Creates a Jetpack task. How the task gets executed is up to
        jetpack runtime.

        Keyword arguments:
        task -- task to create
        """

        request = self._build_create_task_request(task, args, kwargs)
        response: remote_pb2.CreateTaskResponse = await self.dial().CreateTask(request)

        task.id = uuid.UUID(response.task_id)
        if task.is_scheduled:
            instrumentation.get_tracer().scheduled_task_created(task)
        else:
            instrumentation.get_tracer().task_created(task)

    async def cancel_task(self, task_id: str) -> bool:
        req = remote_pb2.CancelTaskRequest(task_id=task_id)
        resp: remote_pb2.CancelTaskResponse = await self.dial().CancelTask(req)
        return resp.success

    async def post_result(
        self, exec_id: str, value: Any = None, error: Optional[Exception] = None
    ) -> remote_pb2.PostResultResponse:

        if not exec_id:
            # Note: value=None and error=None is acceptable because a job
            # can run successfully and return nothing.
            raise Exception("An exec_id is required to post a result")

        result = remote_pb2.Result()
        if error:
            result.error.code = remote_pb2.APPLICATION
            result.error.message = str(error)

            result.error.encoded_error = jetpickle.encode(error)
        else:  # order matters, as 'value' can be None
            result.value.encoded_value = jetpickle.encode(value)

        request = remote_pb2.PostResultRequest(
            exec_id=exec_id,
            result=result,
        )

        response = await self.dial().PostResult(request)
        return cast(remote_pb2.PostResultResponse, response)

    async def wait_for_result(self, task_id: uuid.UUID, symbol_name: str) -> Any:
        request = remote_pb2.WaitForResultRequest(task_id=str(task_id))
        response = await self.dial().WaitForResult(request)
        instrumentation.get_tracer().result_returned(task_id, response.result)
        return self._transform_response_exception(response, symbol_name)

    async def get_task(self, task_id: str) -> remote_pb2.GetTaskResponse:
        request = remote_pb2.GetTaskRequest(task_id=task_id)
        response: remote_pb2.GetTaskResponse = await self.dial().GetTask(request)
        return response

    @staticmethod
    def _build_create_task_request(
        task: Task,
        args: Optional[Tuple[Any, ...]],
        kwargs: Optional[Dict[str, Any]],
    ) -> remote_pb2.CreateTaskRequest:
        encoded_args = b""
        if args or kwargs:
            encoded_args = codec.encode_args(
                args if args else None,
                kwargs if kwargs else None,
            ).encode("utf-8")

        if len(encoded_args) > MAX_MESSAGE_LENGTH:
            raise RuntimeError(
                f"Function argument size is over {math.floor(MAX_MESSAGE_LENGTH / 1000000)}MB"
            )

        current_namespace = namespace.get()
        task_proto_obj = remote_pb2.Task(
            qualified_symbol=task.jetpack_function.name(),
            encoded_args=encoded_args,
            hostname=os.environ.get("HOSTNAME", ""),  # k8s sets this
            app_name=os.environ.get("JETPACK_LABEL_APP_INSTANCE", ""),
            chart_revision=os.environ.get("JETPACK_LABEL_CHART_REVISION", ""),
            parent_task_id=task.parent_task_id,
            with_checkpointing=task.with_checkpointing,
            target_time=Timestamp(seconds=task.target_time),
        )
        if current_namespace:
            task_proto_obj.namespace = current_namespace

        return remote_pb2.CreateTaskRequest(
            task=task_proto_obj, sdk_version=__version__
        )

    @staticmethod
    def _transform_response_exception(
        response: remote_pb2.WaitForResultResponse,
        symbol_name: str,
    ) -> Any:
        if response.result.HasField("error") and response.result.error.encoded_error:
            err = jetpickle.decode(response.result.error.encoded_error)
            jetpack_err = JetpackError(f"chained exception for: {symbol_name}")
            raise jetpack_err from err
        elif response.result.HasField("value"):
            val = jetpickle.decode(response.result.value.encoded_value)
            return val
        else:
            raise RuntimeError(
                f"Either 'value' or 'error' should be specified in response. Got: {response}"
            )


client = Client()


def init(host: str, api_key: str) -> None:
    client.set_runtime_host(host, api_key)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: models/core/core.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='models/core/core.proto',
  package='core',
  syntax='proto3',
  serialized_options=b'Z\037go.jetpack.io/axiom/models/core',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16models/core/core.proto\x12\x04\x63ore\x1a\x1fgoogle/protobuf/timestamp.proto\"v\n\x08NodeMeta\x12\n\n\x02id\x18\x01 \x01(\t\x12.\n\ncreated_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nupdated_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x1d\n\x0bPointerMeta\x12\x0e\n\x06loaded\x18\x01 \x01(\x08\"\x0b\n\tNonDBMetaB!Z\x1fgo.jetpack.io/axiom/models/coreb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])




_NODEMETA = _descriptor.Descriptor(
  name='NodeMeta',
  full_name='core.NodeMeta',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='core.NodeMeta.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_at', full_name='core.NodeMeta.created_at', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='updated_at', full_name='core.NodeMeta.updated_at', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=65,
  serialized_end=183,
)


_POINTERMETA = _descriptor.Descriptor(
  name='PointerMeta',
  full_name='core.PointerMeta',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='loaded', full_name='core.PointerMeta.loaded', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=185,
  serialized_end=214,
)


_NONDBMETA = _descriptor.Descriptor(
  name='NonDBMeta',
  full_name='core.NonDBMeta',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=216,
  serialized_end=227,
)

_NODEMETA.fields_by_name['created_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_NODEMETA.fields_by_name['updated_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
DESCRIPTOR.message_types_by_name['NodeMeta'] = _NODEMETA
DESCRIPTOR.message_types_by_name['PointerMeta'] = _POINTERMETA
DESCRIPTOR.message_types_by_name['NonDBMeta'] = _NONDBMETA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

NodeMeta = _reflection.GeneratedProtocolMessageType('NodeMeta', (_message.Message,), {
  'DESCRIPTOR' : _NODEMETA,
  '__module__' : 'models.core.core_pb2'
  # @@protoc_insertion_point(class_scope:core.NodeMeta)
  })
_sym_db.RegisterMessage(NodeMeta)

PointerMeta = _reflection.GeneratedProtocolMessageType('PointerMeta', (_message.Message,), {
  'DESCRIPTOR' : _POINTERMETA,
  '__module__' : 'models.core.core_pb2'
  # @@protoc_insertion_point(class_scope:core.PointerMeta)
  })
_sym_db.RegisterMessage(PointerMeta)

NonDBMeta = _reflection.GeneratedProtocolMessageType('NonDBMeta', (_message.Message,), {
  'DESCRIPTOR' : _NONDBMETA,
  '__module__' : 'models.core.core_pb2'
  # @@protoc_insertion_point(class_scope:core.NonDBMeta)
  })
_sym_db.RegisterMessage(NonDBMeta)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import asyncio

import click

from jetpack import cron
from jetpack._task.jetpack_function import JetpackFunction
from jetpack.cmd import util
from jetpack.cmd.params import ENTRYPOINT_PARAMS
from jetpack.config import symbols
from jetpack.config.symbols import Symbol


@click.group(name="cron", help="Cronjob specific commands")
def cron_group() -> None:
    pass


@click.command(help="List existing cronjobs")
@click.option("--entrypoint", **ENTRYPOINT_PARAMS)
def ls(entrypoint: str) -> None:
    util.load_user_entrypoint(entrypoint)
    cronjobs = cron.get_jobs()
    if not cronjobs:
        click.echo("No cronjobs found")
        return
    click.echo(f"Found {len(cronjobs)} cronjobs:")
    for cronjob in cronjobs:
        click.echo("---")
        click.echo(cron.pretty_print(cronjob))


@click.command(help="Executes specified cronjob")
@click.option("--entrypoint", **ENTRYPOINT_PARAMS)
@click.argument("symbol-name")
def exec(entrypoint: str, symbol_name: str) -> None:
    util.load_user_entrypoint(entrypoint)
    func = symbols.get_symbol_table()[Symbol(symbol_name)]
    _, err = asyncio.run(JetpackFunction(func).exec(post_result=False))
    if err:
        raise err


cron_group.add_command(ls)
cron_group.add_command(exec)



from typing import Any, Protocol, cast

import jsonpickle
import tblib

from jetpack.config import symbols


def encode(val: Any) -> bytes:
    encoded = val
    if isinstance(val, BaseException):
        encoded = _pickle_prep_exception(val)
    return bytes(jsonpickle.encode(encoded), "utf-8")


def decode(val: bytes) -> Any:

    symbols.get_symbol_table().enable_key_overwrite(True)
    decoded = jsonpickle.decode(val)
    symbols.get_symbol_table().enable_key_overwrite(False)

    if isinstance(decoded, BaseException):
        decoded = _unpickle_prep_exception(decoded)
    return decoded


class ExceptionWithJetpackFields(BaseException):
    jetpack_cause: Any
    jetpack_context: Any
    jetpack_traceback: Any
    __cause__: Any
    __context__: Any
    __traceback__: Any


# pickle_prep_exception prepares the python exception object for json_pickling. It does
# this by converting the native C-object at exception.__traceback__ to be a
# python dict. It does this recursively by following the exception chain via
# exception.__cause__ and exception.__context__.
#
# The "prepared exception" objects are stored at `.jetpack_traceback`,
# `.jetpack_cause` and `jetpack_context` fields.
def _pickle_prep_exception(err: BaseException) -> BaseException:
    error = cast(ExceptionWithJetpackFields, err)

    error.jetpack_traceback = tblib.Traceback(error.__traceback__).to_dict()

    error.jetpack_cause = None
    if error.__cause__ is not None:
        error.jetpack_cause = _pickle_prep_exception(error.__cause__)

    error.jetpack_context = None
    if error.__context__ is not None:
        error.jetpack_context = _pickle_prep_exception(error.__context__)

    return error


# unpickle_prep_exception undoes the changes in pickle_prep_exception, and sets the
# jetpack-specific fields on the exception object to None
def _unpickle_prep_exception(err: BaseException) -> BaseException:
    error = cast(ExceptionWithJetpackFields, err)

    if error.jetpack_traceback is not None:
        error.__traceback__ = tblib.Traceback.from_dict(
            error.jetpack_traceback
        ).as_traceback()

    if error.jetpack_cause is not None:
        error.__cause__ = _unpickle_prep_exception(error.jetpack_cause)

    if error.jetpack_context is not None:
        error.__context__ = _unpickle_prep_exception(error.jetpack_context)

    error.jetpack_cause = None
    error.jetpack_context = None
    error.jetpack_traceback = None

    return error



import inspect
from typing import Any, Awaitable, Callable, Optional, TypeVar, Union, cast

from deprecation import deprecated

from jetpack import utils
from jetpack._task.errors import NotAsyncError
from jetpack._task.jetpack_function import JetpackFunction
from jetpack._task.jetpack_function_with_client import schedule as schedule
from jetpack.config import symbols

T = TypeVar("T")
__pdoc__ = {}
__pdoc__["jet"] = "Alias for function"
__pdoc__["jetroutine"] = "Alias for function"


# @function is our general remote work decorator. It does not specify how the
# work will be done (RPC, job, queue, etc) and instead leaves that as an
# implementation detail.
def jetroutine_decorator(
    # The stararg forces the arguments after that to be specified with the keyword
    #
    # > Parameters after “*” [...] are keyword-only parameters and may only be passed by keyword arguments
    # https://docs.python.org/3/reference/compound_stmts.html#function-definitions
    fn: Optional[Callable[..., T]] = None,
    *,
    with_checkpointing: bool = False,
) -> Union[
    Callable[..., Awaitable[T]],
    Callable[[Callable[..., T]], Callable[..., Awaitable[T]]],
]:
    """Decorator that wraps any async Python function, and returns a Jetroutine

    Async python functions decorated with @function (or @jet or @jetroutine) will
    be wrapped as JetpackFunctions, and registered with the runtime. Calling these
    functions will run them as remote Jetroutines, instead of running them locally

    Returns:
        `jetpack._task.jetpack_function.JetpackFunction[T]`

    """
    def wrapper(fn: Callable[..., T]) -> Callable[..., Awaitable[T]]:
        # Use asyncio.iscoroutine() instead?
        if not inspect.iscoroutinefunction(fn):
            raise NotAsyncError(
                f"Jetpack functions must be async. {utils.qualified_func_name(fn)} is not async."
            )
        symbols.get_symbol_table().register(fn)
        task: JetpackFunction[T] = JetpackFunction(fn, with_checkpointing)
        return task

    return wrapper(fn) if fn else wrapper


@deprecated(details="Use jetroutine instead.")
def function(
    fn: Optional[Callable[..., T]] = None, *, with_checkpointing: bool = False
) -> Union[
    Callable[..., Awaitable[T]],
    Callable[[Callable[..., T]], Callable[..., Awaitable[T]]],
]:
    # can enable this print to be more aggressive, since smarthop and other
    # customers may not have python's dev-mode on to see the @deprecated decorator's
    # warning message:
    #
    # print("WARNING: @function is deprecated. Use @jetroutine instead.")

    return jetroutine_decorator(fn, with_checkpointing=with_checkpointing)


@deprecated(details="Use jetroutine instead.")
def jet(
    fn: Optional[Callable[..., T]] = None, *, with_checkpointing: bool = False
) -> Union[
    Callable[..., Awaitable[T]],
    Callable[[Callable[..., T]], Callable[..., Awaitable[T]]],
]:
    return jetroutine_decorator(fn, with_checkpointing=with_checkpointing)


jetroutine = jetroutine_decorator


def workflow(
    fn: Optional[Callable[..., T]] = None
) -> Union[
    Callable[..., Awaitable[T]],
    Callable[[Callable[..., T]], Callable[..., Awaitable[T]]],
]:
    return jetroutine_decorator(fn, with_checkpointing=True)
